## Alarm Server
## Supporting Envisalink 2DS/3
##
## This code is under the terms of the GPL v3 license.
import ctypes
c_uint16 = ctypes.c_uint16

class IconLED_Bitfield( ctypes.LittleEndianStructure ):
    _fields_ = [
                ("alarm",     c_uint16, 1 ),
                ("alarm_in_memory", c_uint16, 1 ),
                ("armed_away",    c_uint16, 1 ),
                ("ac_present",       c_uint16, 1 ),
                ("bypass",       c_uint16, 1 ),
                ("chime",       c_uint16, 1 ),
                ("not_used1",       c_uint16, 1 ),
                ("armed_zero_entry_delay",       c_uint16, 1 ),
                ("alarm_fire_zone",       c_uint16, 1 ),
                ("system_trouble",       c_uint16, 1 ),
                ("not_used2",       c_uint16, 1 ),
                ("not_used3",       c_uint16, 1 ),
                ("ready",       c_uint16, 1 ),
                ("fire",       c_uint16, 1 ),
                ("low_battery",       c_uint16, 1 ),
                ("armed_stay",       c_uint16, 1 )
               ]

class IconLED_Flags( ctypes.Union ):
    _fields_ = [
                ("b",      IconLED_Bitfield ),
                ("asShort", c_uint16    )
               ]
    _anonymous_ = ("b")

evl_Commands = {
    'KeepAlive' : '00',
    'ChangeDefaultPartition' : '01',
    'DumpZoneTimers' : '02',
    'PartitionKeypress' : '03'
}

evl_PanicTypes = {
    'Fire' : 'A',
    'Ambulance' : 'B',
    'Police' : 'C'
}

evl_ResponseTypes = {
    'Login:' :  {'name' : 'Login Prompt', 'description' : 'Sent During Session Login Only.', 'handler' : 'login'},
    'OK' : {'name' : 'Login Success', 'description' : 'Send During Session Login Only, successful login', 'handler' : 'login_success'},
    'FAILED' : {'name' : 'Login Failure', 'description' : 'Sent During Session Login Only, password not accepted', 'handler' : 'login_failure'},
    'Timed Out!' : {'name' : 'Login Interaction Timed Out', 'description' : 'Sent during Session Login Only, socket connection is then closed', 'handler' : 'login_timeout'},
    '%00' : {'name' : 'Virtual Keypad Update', 'description' : 'The panel wants to update the state of the keypad','handler' : 'keypad_update'},
    '%01' : {'type' : 'zone', 'name' : 'Zone State Change', 'description' : 'A zone change-of-state has occurred', 'handler' : 'zone_state_change'},
    '%02' : {'type' : 'partition', 'name' : 'Partition State Change', 'description' : 'A partition change-of-state has occured', 'handler' : 'partition_state_change'},
    '%03' : {'type' : 'system', 'name' : 'Realtime CID Event', 'description' : 'A system event has happened that is signaled to either the Envisalerts servers or the central monitoring station', 'handler' : 'realtime_cid_event'},
    '%FF' : {'name' : 'Envisalink Zone Timer Dump', 'description' : 'This command contains the raw zone timers used inside the Envisalink. The dump is a 256 character packed HEX string representing 64 UINT16 (little endian) zone timers. Zone timers count down from 0xFFFF (zone is open) to 0x0000 (zone is closed too long ago to remember). Each ''tick'' of the zone time is actually 5 seconds so a zone timer of 0xFFFE means ''5 seconds ago''. Remember, the zone timers are LITTLE ENDIAN so the above example would be transmitted as FEFF.', 'handler' : 'zone_timer_dump'},
    '^00' : {'type' : 'envisalink', 'name': 'Poll', 'description' : 'Envisalink poll', 'handler' : 'poll_response'},
    '^01' : {'type' : 'envisalink', 'name': 'Change Default Partition', 'description': 'Change the partition which keystrokes are sent to when using the virtual keypad.', 'handler' : 'command_response'},
    '^02' : {'type' : 'envisalink', 'name': 'Dump Zone Timers', 'description' : 'This command contains the raw zone timers used inside the Envisalink. The dump is a 256 character packed HEX string representing 64 UINT16 (little endian) zone timers. Zone timers count down from 0xFFFF (zone is open) to 0x0000 (zone is closed too long ago to remember). Each ''tick'' of the zone time is actually 5 seconds so a zone timer of 0xFFFE means ''5 seconds ago''. Remember, the zone timers are LITTLE ENDIAN so the above example would be transmitted as FEFF.','handler' : 'command_response'},
    '^03' : {'type' : 'envisalink', 'name': 'Keypress to Specific Partition', 'description' : 'This will send a keystroke to the panel from an arbitrary partition. Use this if you don''t want to change the TPI default partition.' ,'handler' : 'command_response'},
    '^0C' : {'type' : 'envisalink', 'name': 'Response for Invalid Command', 'description' : 'This response is returned when an invalid command number is passed to Envisalink', 'handler': 'command_response'}
}

evl_TPI_Response_Codes = {
    '00' : 'Command Accepted',
    '01' : 'Receive Buffer Overrun (a command is received while another is still being processed)',
    '02' : 'Unknown Command',
    '03' : 'Syntax Error. Data appended to the command is incorrect in some fashion',
    '04' : 'Receive Buffer Overflow',
    '05' : 'Receive State Machine Timeout (command not completed within 3 seconds)'
}
evl_Partition_Status_Codes = {
    '00' : {'name' : 'NOT_USED', 'description' : 'Partition is not used or doesn''t exist'},
    '01' : {'name' : 'READY', 'description' : 'Ready', 'pluginhandler' : 'disarmed'},
    '02' : {'name' : 'READY_BYPASS', 'description' : 'Ready to Arm (Zones are Bypasses)', 'pluginhandler' : 'disarmed'},
    '03' : {'name' : 'NOT_READY', 'description' : 'Not Ready', 'pluginhandler' : 'disarmed'},
    '04' : {'name' : 'ARMED_STAY', 'description' : 'Armed in Stay Mode', 'pluginhandler' : 'armedHome'},
    '05' : {'name' : 'ARMED_AWAY', 'description' : 'Armed in Away Mode', 'pluginhandler' : 'armedAway'},
    '06' : {'name' : 'ARMED_MAX', 'description' : 'Armed in Away Mode', 'pluginhandler' : 'armedInstant'},
    '07' : {'name' : 'EXIT_ENTRY_DELAY', 'description' : 'Entry or Exit Delay'},
    '08' : {'name' : 'IN_ALARM', 'description' : 'Partition is in Alarm', 'pluginhandler' : 'alarmTriggered'},
    '09' : {'name' : 'ALARM_IN_MEMORY', 'description' : 'Alarm Has Occurred (Alarm in Memory)', 'pluginhandler' : 'alarmCleared'}
}

evl_Virtual_Keypad_How_To_Beep = {
    '00' : 'off',
    '01' : 'beep 1 time',
    '02' : 'beep 2 times',
    '03' : 'beep 3 times',
    '04' : 'continous fast beep',
    '05' : 'continuous slow beep'
}

evl_CID_Qualifiers = {
    1 : 'New Event or Opening',
    3 : 'New Restore or Closing',
    6 : 'Previously Reported Condition Still Present'
}

evl_ArmDisarm_CIDs = [401, 403, 407, 408, 409, 441, 442]

evl_CID_Events = {
    100 : {
    "label" : "Medical Alert",
    "type"  : "zone",
    },
    101 : {
    "label" : "Personal Emergency",
    "type"  : "zone",
    },
    102 : {
    "label" : "Failure to Report In",
    "type"  : "zone",
    },
    110 : {
    "label" : "Fire Alarm",
    "type"  : "zone",
    },
    111 : {
    "label" : "Smoke Alarm",
    "type"  : "zone",
    },
    112 : {
    "label" : "Combustion Detected Alarm",
    "type"  : "zone",
    },
    113 : {
    "label" : "Water Flood Alarm",
    "type"  : "zone",
    },
    114 : {
    "label" : "Excessive Heat Alarm",
    "type"  : "zone",
    },
    115 : {
    "label" : "Fire Alarm Pulled",
    "type"  : "zone",
    },
    116 : {
    "label" : "Duct Alarm",
    "type"  : "zone",
    },
    117 : {
    "label" : "Flame Detected",
    "type"  : "zone",
    },
    118 : {
    "label" : "Near Alarm",
    "type"  : "zone",
    },
    120 : {
    "label" : "Panic Alarm",
    "type"  : "zone",
    },
    121 : {
    "label" : "Duress Alarm",
    "type"  : "user",
    },
    122 : {
    "label" : "Alarm, 24-hour Silent",
    "type"  : "zone",
    },
    123 : {
    "label" : "Alarm, 24-hour Audible",
    "type"  : "zone",
    },
    124 : {
    "label" : "Duress - Access granted",
    "type"  : "zone",
    },
    125 : {
    "label" : "Duress - Egress granted",
    "type"  : "zone",
    },
    130 : {
    "label" : "Burgalry in Progress",
    "type"  : "zone",
    },
    131 : {
    "label" : "Alarm, Perimeter",
    "type"  : "zone",
    },
    132 : {
    "label" : "Alarm, Interior",
    "type"  : "zone",
    },
    133 : {
    "label" : "24 Hour (Safe)",
    "type"  : "zone",
    },
    134 : {
    "label" : "Alarm, Entry/Exit",
    "type"  : "zone",
    },
    135 : {
    "label" : "Alarm, Day/Night",
    "type"  : "zone",
    },
    136 : {
    "label" : "Alarm, Outdoor",
    "type"  : "zone",
    },
    137 : {
    "label" : "Alarm, Tamper",
    "type"  : "zone",
    },
    138 : {
    "label" : "Near Alarm",
    "type"  : "zone",
    },
    139 : {
    "label" : "Intrusion Verifier",
    "type"  : "zone",
    },
    140 : {
    "label" : "Alarm, General Alarm",
    "type"  : "zone",
    },
    141 : {
    "label" : "Alarm, Polling Loop Open",
    "type"  : "zone",
    },
    142 : {
    "label" : "Alarm, Polling Loop Short",
    "type"  : "zone",
    },
    143 : {
    "label" : "Alarm, Expansion Module",
    "type"  : "zone",
    },
    144 : {
    "label" : "Alarm, Sensor Tamper",
    "type"  : "zone",
    },
    145 : {
    "label" : "Alarm, Expansion Module Tamper",
    "type"  : "zone",
    },
    146 : {
    "label" : "Silent Burglary",
    "type"  : "zone",
    },
    147 : {
    "label" : "Sensor Supervision failure",
    "type"  : "zone",
    },
    150 : {
    "label" : "Alarm, 24-Hour Auxiliary",
    "type"  : "zone",
    },
    151 : {
    "label" : "Alarm, Gas detected",
    "type"  : "zone",
    },
    152 : {
    "label" : "Alarm, Refrigeration",
    "type"  : "zone",
    },
    153 : {
    "label" : "Alarm, Loss of heat",
    "type"  : "zone",
    },
    154 : {
    "label" : "Alarm, Water leakage",
    "type"  : "zone",
    },
    155 : {
    "label" : "Alarm, foil break",
    "type"  : "zone",
    },
    156 : {
    "label" : "Day trouble",
    "type"  : "zone",
    },
    157 : {
    "label" : "Low bottled gas level",
    "type"  : "zone",
    },
    158 : {
    "label" : "Alarm, High temperature",
    "type"  : "zone",
    },
    159 : {
    "label" : "Alarm, Low temperature",
    "type"  : "zone",
    },
    161 : {
    "label" : "Alarm, Loss of air flow",
    "type"  : "zone",
    },
    162 : {
    "label" : "Alarm, Carbon Monoxide Detected",
    "type"  : "zone",
    },
    163 : {
    "label" : "Alarm, Tank Level",
    "type"  : "zone",
    },

    300 : {
    "label" : "System Trouble",
    "type"  : "zone",
    },
    301 : {
    "label" : "AC Power",
    "type"  : "zone",
    },
    302 : {
    "label" : "Low System Battery/Battery Test Fail",
    "type"  : "zone",
    },
    303 : {
    "label" : "RAM Checksum Bad",
    "type"  : "zone",
    },
    304 : {
    "label" : "ROM Checksum Bad",
    "type"  : "zone",
    },
    305 : {
    "label" : "System Reset",
    "type"  : "zone",
    },
    306 : {
    "label" : "Panel programming changed",
    "type"  : "zone",
    },
    307 : {
    "label" : "Self-test failure",
    "type"  : "zone",
    },
    308 : {
    "label" : "System shutdown",
    "type"  : "zone",
    },
    309 : {
    "label" : "Battery test failure",
    "type"  : "zone",
    },
    310 : {
    "label" : "Ground fault",
    "type"  : "zone",
    },
    311 : {
    "label" : "Battery Missing/Dead",
    "type"  : "zone",
    },
    312 : {
    "label" : "Power Supply Overcurrent",
    "type"  : "zone",
    },
    313 : {
    "label" : "Engineer Reset",
    "type"  : "user",
    },
    321 : {
    "label" : "Bell/Siren Trouble",
    "type"  : "zone",
    },
    333 : {
    "label" : "Trouble or Tamper Expansion Module",
    "type"  : "zone",
    },
    341 : {
    "label" : "Trouble, ECP Cover Tamper",
    "type"  : "zone",
    },
    344 : {
    "label" : "RF Receiver Jam",
    "type"  : "zone",
    },
    351 : {
    "label" : "Telco Line Fault",
    "type"  : "zone",
    },
    353 : {
    "label" : "Long Range Radio Trouble",
    "type"  : "zone",
    },
    373 : {
    "label" : "Fire Loop Trouble",
    "type"  : "zone",
    },
    374 : {
    "label" : "Exit Error Alarm",
    "type"  : "zone",
    },
    380 : {
    "label" : "Global Trouble, Trouble Day/Night",
    "type"  : "zone",
    },
    381 : {
    "label" : "RF Supervision Trouble",
    "type"  : "zone",
    },
    382 : {
    "label" : "Supervision Auxillary Wire Zone",
    "type"  : "zone",
    },
    383 : {
    "label" : "RF Sensor Tamper",
    "type"  : "zone",
    },
    384 : {
    "label" : "RF Sensor Low Battery",
    "type"  : "zone",
    },
    393 : {
    "label" : "Clean Me",
    "type"  : "zone",
    },

    401 : {
    "label" : "AWAY/MAX",
    "type"  : "user",
    },
    403 : {
    "label" : "Scheduled Arming",
    "type"  : "user",
    },
    406 : {
    "label" : "Cancel by User",
    "type"  : "user",
    },
    407 : {
    "label" : "Remote Arm/Disarm (Downloading)",
    "type"  : "user",
    },
    408 : {
    "label" : "Quick AWAY/MAX",
    "type"  : "user",
    },
    409 : {
    "label" : "AWAY/MAX Keyswitch",
    "type"  : "user",
    },
    411 : {
    "label" : "Callback Requested",
    "type"  : "user",
    },
    412 : {
    "label" : "Success-Download/Access",
    "type"  : "user",
    },
    413 : {
    "label" : "Unsuccessful Access",
    "type"  : "user",
    },
    414 : {
    "label" : "System Shutdown",
    "type"  : "user",
    },
    415 : {
    "label" : "Dialer Shutdown",
    "type"  : "user",
    },
    416 : {
    "label" : "Successful Upload",
    "type"  : "user",
    },
    421 : {
    "label" : "Access Denied",
    "type"  : "user",
    },
    422 : {
    "label" : "Access Granted",
    "type"  : "user",
    },
    423 : {
    "label" : "PANIC Forced Access",
    "type"  : "zone",
    },
    424 : {
    "label" : "Egress Denied",
    "type"  : "user",
    },
    425 : {
    "label" : "Egress Granted",
    "type"  : "user",
    },
    426 : {
    "label" : "Access Door Propped Open",
    "type"  : "zone",
    },
    427 : {
    "label" : "Access Point DSM Trouble",
    "type"  : "zone",
    },
    428 : {
    "label" : "Access Point RTE Trouble",
    "type"  : "zone",
    },
    429 : {
    "label" : "Access Program Mode Entry",
    "type"  : "user",
    },
    430 : {
    "label" : "Access Program Mode Exit",
    "type"  : "user",
    },
    431 : {
    "label" : "Access Threat Level Change",
    "type"  : "user",
    },
    432 : {
    "label" : "Access Relay/Triger Failure",
    "type"  : "zone",
    },
    433 : {
    "label" : "Access RTE Shunt",
    "type"  : "zone",
    },
    434 : {
    "label" : "Access DSM Shunt",
    "type"  : "zone",
    },
    441 : {
    "label" : "STAY/INSTANT",
    "type"  : "user",
    },
    442 : {
    "label" : "STAY/INSTANT Keyswitch",
    "type"  : "user",
    },

    570 : {
    "label" : "Zone Bypass",
    "type"  : "zone",
    },

    574 : {
    "label" : "Group Bypass",
    "type"  : "user"
    },

    601 : {
    "label" : "Operator Initiated Dialer Test",
    "type"  : "user",
    },
    602 : {
    "label" : "Periodic Test",
    "type"  : "zone",
    },
    606 : {
    "label" : "AAV to follow",
    "type"  : "zone",
    },
    607 : {
    "label" : "Walk Test",
    "type"  : "user",
    },
    623 : {
    "label" : "Event Log 80% Full",
    "type"  : "zone",
    },
    625 : {
    "label" : "Real-Time Clock Changed",
    "type"  : "user",
    },
    627 : {
    "label" : "Program Mode Entry",
    "type"  : "zone",
    },
    628 : {
    "label" : "Program Mode Exit",
    "type"  : "zone",
    },
    629 : {
    "label" : "1-1/3 Day No Event",
    "type"  : "zone",
    },
    642 : {
    "label" : "Latch Key",
    "type"  : "user",
    },
    750 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    751 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    752 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    753 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    754 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    755 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    756 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    757 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    758 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    759 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    760 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    761 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    762 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    763 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    764 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    765 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    766 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    767 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    768 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    769 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    770 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    771 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    772 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    773 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    774 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    775 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    776 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    777 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    778 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    779 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    780 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    781 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    782 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    783 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    784 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    785 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    786 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    787 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    788 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
    789 : {
    "label" : "Configurable Zone Type",
    "type"  : "zone",
    },
}



class AlarmState:
    """Helper class for alarm state functionality."""

    @staticmethod
    def get_initial_alarm_state(maxZones, maxPartitions):
        """Builds the proper alarm state collection."""

        _alarmState = {'partition': {}, 'zone': {}}

        for i in range(1, maxPartitions + 1):
            _alarmState['partition'][i] = {'status': {'partition_state': 'N/A', 
                                                      'alpha': 'N/A', 
                                                      'ac_present': True, 
                                                      'beep': False, 
                                                      'bypass': False, 
                                                      'chime': False, 
                                                      'entry_delay': False, 
                                                      'exit_delay': False, 
                                                      'last_armed_by_user': '', 
                                                      'last_disarmed_by_user': '', 
                                                      'ready': False, 
                                                      'bat_trouble': False, 
                                                      'trouble': False, 
                                                      'fire': False, 
                                                      'alarm': False, 
                                                      'alarm_fire_zone': False, 
                                                      'alarm_in_memory': False, 
                                                      'armed_away': False, 
                                                      'armed_stay': False, 
                                                      'armed_zero_entry_delay': False }}
        for j in range (1, maxZones + 1):
            _alarmState['zone'][j] = {'status': {'open': False, 'fault': False, 'alarm': False, 'tamper': False}, 'last_fault': 0, 'bypassed': False}

        return _alarmState



#!/usr/bin/env python3
import signal
import sys
from pyenvisalink import EnvisalinkAlarmPanel

#This is a test harness for the pyenvisalink library.  It will assist in testing the library against both Honeywell and DSC.

#Get Details from the user...
ip = input("Please input the IP address of your envisalink device: ")
port = input("Please input the port of your envisalink device (4025 is default): ")
version = input("Which envisalink device do you have? Enter 3 for evl3 or 4 for evl4: ")
panel = input("Input DSC if you have a DSC panel, or HONEYWELL if you have a honeywell panel: ")
pw = input("Please input your envisalink password: ")

na = input("Config complete. Please press enter now to connect to the envisalink.  When finished, use Ctrl+C to disconnect and exit")
testpanel = EnvisalinkAlarmPanel(ip, int(port), panel, int(version), pw, pw)
testpanel.start()

def signal_handler(signal, frame):
        print('You pressed Ctrl+C!')
        testpanel.stop()
        sys.exit(0)

signal.signal(signal.SIGINT, signal_handler)
signal.pause()



import asyncio
import logging
import json
import re
import asyncio
import datetime
from pyenvisalink import EnvisalinkClient
from pyenvisalink.dsc_envisalinkdefs import *

_LOGGER = logging.getLogger(__name__)

from asyncio import ensure_future

class DSCClient(EnvisalinkClient):
    """Represents a dsc alarm client."""

    def __init__(self, panel, loop):
        self._refreshZoneByassState = False
        self._zoneBypassRefreshTask = None
        super().__init__(panel, loop)

    def to_chars(self, string):
        chars = []
        for char in string:
            chars.append(ord(char))
        return chars

    def get_checksum(self, code, data):
        """part of each command includes a checksum.  Calculate."""
        return ("%02X" % sum(self.to_chars(code)+self.to_chars(data)))[-2:]

    def send_command(self, code, data):
        """Send a command in the proper honeywell format."""
        to_send = code + data + self.get_checksum(code, data)
        self.send_data(to_send)

    def dump_zone_timers(self):
        """Send a command to dump out the zone timers."""
        self.send_command(evl_Commands['DumpZoneTimers'], '')

    def keypresses_to_partition(self, partitionNumber, keypresses):
        """Send keypresses (max of 6) to a particular partition."""
        self.send_command(evl_Commands['PartitionKeypress'], str.format("{0}{1}", partitionNumber, keypresses[:6]))

    async def keep_alive(self):
        """Send a keepalive command to reset it's watchdog timer."""
        while not self._shutdown:
            if self._loggedin:
                self.send_command(evl_Commands['KeepAlive'], '')
            await asyncio.sleep(self._alarmPanel.keepalive_interval)

    async def periodic_zone_timer_dump(self):
        """Used to periodically get the zone timers to make sure our zones are updated."""
        while not self._shutdown:
            if self._loggedin:
                self.dump_zone_timers()
            await asyncio.sleep(self._alarmPanel.zone_timer_interval)

    def arm_stay_partition(self, code, partitionNumber):
        """Public method to arm/stay a partition."""
        self._cachedCode = code
        self.send_command(evl_Commands['ArmStay'], str(partitionNumber))

    def arm_away_partition(self, code, partitionNumber):
        """Public method to arm/away a partition."""
        self._cachedCode = code
        self.send_command(evl_Commands['ArmAway'], str(partitionNumber))

    def arm_max_partition(self, code, partitionNumber):
        """Public method to arm/max a partition."""
        self._cachedCode = code
        self.send_command(evl_Commands['ArmMax'], str(partitionNumber))

    def arm_night_partition(self, code, partitionNumber):
        """Public method to arm/max a partition."""
        self.arm_max_partition(code, partitionNumber)

    def disarm_partition(self, code, partitionNumber):
        """Public method to disarm a partition."""
        self._cachedCode = code
        self.send_command(evl_Commands['Disarm'], str(partitionNumber) + str(code))

    def panic_alarm(self, panicType):
        """Public method to raise a panic alarm."""
        self.send_command(evl_Commands['Panic'], evl_PanicTypes[panicType])

    def toggle_zone_bypass(self, zone):
        """Public method to toggle a zone's bypass state."""
        self.keypresses_to_partition(1, "*1%02d#" % zone)

    def command_output(self, code, partitionNumber, outputNumber):
        """Used to activate the selected command output"""
        self._cachedCode = code
        self.send_command(evl_Commands['CommandOutput'], str.format("{0}{1}", partitionNumber, outputNumber))	

    def parseHandler(self, rawInput):
        """When the envisalink contacts us- parse out which command and data."""
        cmd = {}
        dataoffset = 0
        if rawInput != '':
            if re.match(r'\d\d:\d\d:\d\d\s', rawInput):
                dataoffset = dataoffset + 9
            code = rawInput[dataoffset:dataoffset+3]
            cmd['code'] = code
            cmd['data'] = rawInput[dataoffset+3:][:-2]
            
            try:
                #Interpret the login command further to see what our handler is.
                if evl_ResponseTypes[code]['handler'] == 'login':
                    if cmd['data'] == '3':
                      handler = 'login'
                    elif cmd['data'] == '2':
                      handler = 'login_timeout'
                    elif cmd['data'] == '1':
                      handler = 'login_success'
                    elif cmd['data'] == '0':
                      handler = 'login_failure'

                    cmd['handler'] = "handle_%s" % handler
                    cmd['callback'] = "callback_%s" % handler

                else:
                    cmd['handler'] = "handle_%s" % evl_ResponseTypes[code]['handler']
                    cmd['callback'] = "callback_%s" % evl_ResponseTypes[code]['handler']
            except KeyError:
                _LOGGER.debug(str.format('No handler defined in config for {0}, skipping...', code))
                
        return cmd

    def handle_login(self, code, data):
        """When the envisalink asks us for our password- send it."""
        self.send_command(evl_Commands['Login'], self._alarmPanel.password) 

    def handle_login_success(self, code, data):
        """Handler for when the envisalink accepts our credentials."""
        super().handle_login_success(code, data)
        dt = datetime.datetime.now().strftime('%H%M%m%d%y')
        self.send_command(evl_Commands['SetTime'], dt)
        self.send_command(evl_Commands['StatusReport'], '')

        if self._alarmPanel._zoneBypassEnabled:
            """ Initiate request for zone bypass information """
            self._refreshZoneBypassStatus = True
            if self._zoneBypassRefreshTask == None:
                self._zoneBypassRefreshTask = ensure_future(self.dump_zone_bypass_status(), loop=self._eventLoop)

    def handle_command_response(self, code, data):
        """Handle the envisalink's initial response to our commands."""
        _LOGGER.debug("DSC ack recieved.")

    def handle_command_response_error(self, code, data):
        """Handle the case where the DSC passes back a checksum failure."""
        _LOGGER.error("The previous command resulted in a checksum failure.")

    def handle_poll_response(self, code, data):
        """Handle the response to our keepalive messages."""
        self.handle_command_response(code, data)

    def handle_zone_state_change(self, code, data):
        """Handle when the envisalink sends us a zone change."""
        """Event 601-610."""
        parse = re.match('^[0-9]{3,4}$', data)
        if parse:
            zoneNumber = int(data[-3:])
            self._alarmPanel.alarm_state['zone'][zoneNumber]['status'].update(evl_ResponseTypes[code]['status'])
            _LOGGER.debug(str.format("(zone {0}) state has updated: {1}", zoneNumber, json.dumps(evl_ResponseTypes[code]['status'])))
            return zoneNumber
        else:
            _LOGGER.error("Invalid data has been passed in the zone update.")

    def handle_partition_state_change(self, code, data):
        """Handle when the envisalink sends us a partition change."""
        """Event 650-674, 652 is an exception, because 2 bytes are passed for partition and zone type."""
        if code == '652':
            parse = re.match('^[0-9]{2}$', data)
            if parse:
                partitionNumber = int(data[0])
                self._alarmPanel.alarm_state['partition'][partitionNumber]['status'].update(evl_ArmModes[data[1]]['status'])
                _LOGGER.debug(str.format("(partition {0}) state has updated: {1}", partitionNumber, json.dumps(evl_ArmModes[data[1]]['status'])))
                return partitionNumber
            else:
                _LOGGER.error("Invalid data has been passed when arming the alarm.") 
        else:
            parse = re.match('^[0-9]+$', data)
            if parse:
                partitionNumber = int(data[0])
                self._alarmPanel.alarm_state['partition'][partitionNumber]['status'].update(evl_ResponseTypes[code]['status'])
                _LOGGER.debug(str.format("(partition {0}) state has updated: {1}", partitionNumber, json.dumps(evl_ResponseTypes[code]['status'])))
                
                '''Log the user who last armed or disarmed the alarm'''
                if code == '700':
                    lastArmedBy = {'last_armed_by_user': int(data[1:5])}
                    self._alarmPanel.alarm_state['partition'][partitionNumber]['status'].update(lastArmedBy)
                elif code == '750':
                    lastDisarmedBy = {'last_disarmed_by_user': int(data[1:5])}
                    self._alarmPanel.alarm_state['partition'][partitionNumber]['status'].update(lastDisarmedBy)

                if code == '655' and self._alarmPanel._zoneBypassEnabled:
                    """Partition was disarmed which means the bypassed zones have likley been reset so force a zone bypass refresh"""
                    self._refreshZoneBypassStatus = True

                return partitionNumber
            else:
                _LOGGER.error("Invalid data has been passed in the parition update.")

    def handle_send_code(self, code, data):
        """The DSC will, depending upon settings, challenge us with the code.  If the user passed it in, we'll send it."""
        if self._cachedCode is None:
            _LOGGER.error("The envisalink asked for a code, but we have no code in our cache.")
        else:
            self.send_command(evl_Commands['SendCode'], self._cachedCode)
            self._cachedCode = None

    def handle_keypad_update(self, code, data):
        """Handle general- non partition based info"""
        if code == '849':
            bits = "{0:016b}".format(int(data,16))
            trouble_description = ""
            ac_present = True
            for i in range(0, 7):
                if bits[15-i] == '1':
                    trouble_description += evl_verboseTrouble[i] + ', '
                    if i == 1:
                        ac_present = False
            new_status = {'alpha':trouble_description.strip(', '), 'ac_present': ac_present}
        else:
            new_status = evl_ResponseTypes[code]['status']
       
        for part in self._alarmPanel.alarm_state['partition']:
            self._alarmPanel.alarm_state['partition'][part]['status'].update(new_status)
        _LOGGER.debug(str.format("(All partitions) state has updated: {0}", json.dumps(new_status)))

    def handle_zone_bypass_update(self, code, data):
        """ Handle zone bypass update triggered when *1 is used on the keypad """
        if not self._alarmPanel._zoneBypassEnabled:
            return

        self._refreshZoneBypassStatus = False
        if len(data) == 16:
            updates = {}
            for byte in range(8):
                bypassBitfield = int('0x' + data[byte * 2] + data[(byte * 2) + 1], 0)

                for bit in range(8):
                    zoneNumber = (byte * 8) + bit + 1
                    bypassed = (bypassBitfield & (1 << bit) != 0)
                    if self._alarmPanel.alarm_state['zone'][zoneNumber]['bypassed'] != bypassed:
                        updates[zoneNumber] = bypassed
                    self._alarmPanel.alarm_state['zone'][zoneNumber]['bypassed'] = bypassed
                    _LOGGER.debug(str.format("(zone {0}) bypass state has updated: {1}", zoneNumber, bypassed))


            _LOGGER.debug(str.format("zone bypass updates: {0}", updates))
            return updates
        else:
            _LOGGER.error(str.format("Invalid data length ({0}) has been received in the bypass update.", len(data)))

    async def dump_zone_bypass_status(self):
        """ Loop requesting zone bypass status until the command succeeds.
            This is necessary to deal with TPI/DSC buffer overrun errors.
            Ideally all commands would be queued with a retry mechanism when BUSY or BUFFER_OVERRUN is received back"""
        while not self._shutdown:
            if self._loggedin and self._refreshZoneBypassStatus and not self.is_any_partition_armed():
                """ Trigger a 616 'Bypassed Zones Bitfield Dump' to initialize the bypass state.
                    There is unfortunately not a specific command to request a zone bypass dump so the *1# keypresses are sent instead.
                    It appears that limitations in the envisalink API (or perhaps the panel itself) makes it impossible for this feature
                    to work if the alarm panel is setup to require a code to bypass zones.

                    Only issue the command if the alarm is not armed since the EVL will not send a 616 when in the armed state.
                    """
                self.keypresses_to_partition(1, "*1#")
            await asyncio.sleep(5)

    def is_any_partition_armed(self) -> bool:
        """ Check if any partition is either armed or arming/disarming """
        for partition, state in self._alarmPanel.alarm_state['partition'].items():
            status = state['status']
            away = status.get('armed_away', False)
            stay = status.get('armed_stay', False)
            exit_delay = status.get('exit_delay', False)
            entry_delay = status.get('entry_delay', False)
            if away or stay or exit_delay or entry_delay:
                _LOGGER.debug(f"is_any_partition_armed: partition {partition} status={status}")
                return True
        _LOGGER.debug("is_any_partition_armed: no partitions are armed")
        return False




## Alarm Server
## Supporting Envisalink 2DS/3
##
## This code is under the terms of the GPL v3 license.

evl_Commands = {
    'KeepAlive' : '000',
    'StatusReport' : '001',
    'DumpZoneTimers' : '008',
    'PartitionKeypress' : '071',
    'Disarm' : '040',
    'ArmStay' : '031',
    'ArmAway' : '030',
    'ArmMax' : '032',
    'Login' : '005',
    'Panic' : '060',
    'SendCode' : '200',
    'CommandOutput' : '020',
    'SetTime' : '010'
}

evl_PanicTypes = {
    'Fire' : '1',
    'Ambulance' : '2',
    'Police' : '3'
}

evl_ArmModes = {
        '0' : {'name' : 'Arm Away', 'status':{'armed_away': True, 'armed_zero_entry_delay': False, 'alpha':'Arm Away', 'exit_delay':False, 'entry_delay': False }},
        '1' : {'name' : 'Arm Stay', 'status':{'armed_stay': True, 'armed_zero_entry_delay': False, 'alpha':'Arm Stay', 'exit_delay':False, 'entry_delay': False }},
        '2' : {'name' : 'Arm Zero Entry Away', 'status':{'armed_away': True, 'armed_zero_entry_delay': True, 'alpha':'Arm Zero Entry Away', 'exit_delay':False, 'entry_delay': False }},
        '3' : {'name' : 'Arm Zero Entry Stay', 'status':{'armed_stay': True, 'armed_zero_entry_delay': True, 'alpha':'Arm Zero Entry Stay', 'exit_delay':False, 'entry_delay': False }}
    }

evl_ResponseTypes = {
    '505' : {'name':'Login Prompt', 'handler':'login'},
    '615' : {'name':'Envisalink Zone Timer Dump', 'handler':'zone_timer_dump'},
    '500' : {'name':'Poll', 'handler':'poll_response'},
    '501' : {'name':'Checksum', 'handler':'command_response_error'},
    '900' : {'name':'EnterCode', 'handler':'send_code'},
    '912' : {'name':'PGMEnterCode', 'handler':'send_code'},

#ZONE UPDATES

    '601' : {'name':'Zone Alarm', 'handler':'zone_state_change', 'status':{'alarm' : True}},
    '602' : {'name':'Zone Alarm Restore', 'handler':'zone_state_change', 'status':{'alarm' : False}},
    '603' : {'name':'Zone Tamper', 'handler':'zone_state_change', 'status':{'tamper' : True}},
    '604' : {'name':'Zone Tamper Restore', 'handler':'zone_state_change', 'status':{'tamper' : False}},
    '605' : {'name':'Zone Fault', 'handler':'zone_state_change', 'status':{'fault' : True}},
    '606' : {'name':'Zone Fault Restore', 'handler':'zone_state_change', 'status':{'fault' : False}},
    '609' : {'name':'Zone Open', 'handler':'zone_state_change', 'status':{'open' : True}},
    '610' : {'name':'Zone Restored', 'handler':'zone_state_change', 'status':{'open' : False}},

#PARTITION UPDATES
    '650' : {'name':'Ready', 'handler':'partition_state_change', 'status':{'ready' : True, 'alpha' : 'Ready'}},
    '651' : {'name':'Not Ready', 'handler':'partition_state_change', 'status':{'ready' : False, 'alpha' : 'Not Ready'}},
    '652' : {'name':'Armed', 'handler':'partition_state_change'},
    '653' : {'name':'Ready - Force Arming Enabled', 'handler':'partition_state_change', 'status':{'ready': True, 'alpha' : 'Ready - Force Arm'}},
    '654' : {'name':'Alarm', 'handler':'partition_state_change', 'status':{'alarm' : True, 'alpha' : 'Alarm'}},
    '655' : {'name':'Disarmed', 'handler':'partition_state_change', 'status' : {'alarm' : False, 'armed_stay' : False, 'armed_zero_entry_delay': False, 'armed_away' : False, 'exit_delay' : False, 'entry_delay' : False, 'alpha' : 'Disarmed'}},
    '656' : {'name':'Exit Delay in Progress', 'handler':'partition_state_change', 'status':{'exit_delay' : True, 'alpha' : 'Exit Delay In Progress'}},
    '657' : {'name':'Entry Delay in Progress', 'handler':'partition_state_change', 'status':{'entry_delay' : True, 'alpha' : 'Entry Delay in Progress'}},
    '663' : {'name':'ChimeOn', 'handler':'partition_state_change', 'status': {'chime': True}},
    '664' : {'name':'ChimeOff', 'handler':'partition_state_change', 'status': {'chime': False}},
    '673' : {'name':'Busy', 'handler':'partition_state_change', 'status': {'alpha': 'Busy'}},
    '700' : {'name':'Armed by user', 'handler':'partition_state_change', 'status':{}},
    '750' : {'name':'Disarmed by user', 'handler':'partition_state_change', 'status' : {'alarm' : False, 'armed_stay' : False, 'armed_away' : False, 'armed_zero_entry_delay': False, 'exit_delay' : False, 'entry_delay' : False, 'alpha' : 'Disarmed'}},
    '751' : {'name':'Disarmed special', 'handler':'partition_state_change', 'status' : {'alarm' : False, 'armed_stay' : False, 'armed_away' : False, 'armed_zero_entry_delay': False, 'exit_delay' : False, 'entry_delay' : False, 'alpha' : 'Disarmed'}},
    '840' : {'name':'Trouble LED', 'handler':'partition_state_change', 'status':{'trouble' : True}},
    '841' : {'name':'Trouble Clear', 'handler':'partition_state_change', 'status':{'trouble' : False, 'ac_present': True}},

#GENERAL UPDATES
    '621' : {'name':'FireAlarmButton', 'handler':'keypad_update', 'status':{'fire' : True, 'alarm': True, 'alpha' : 'Fire Alarm'}},
    '622' : {'name':'FireAlarmButtonOff', 'handler':'keypad_update', 'status':{'fire' : False, 'alarm': False, 'alpha' : 'Fire Alarm Cleared'}},
    '623' : {'name':'AuxAlarmButton', 'handler':'keypad_update', 'status':{'alarm': True, 'alpha' : 'Aux Alarm'}},
    '624' : {'name':'AuxAlarmButtonOff', 'handler':'keypad_update', 'status':{'alarm': False, 'alpha' : 'Aux Alarm Cleared'}},
    '625' : {'name':'PanicAlarmButton', 'handler':'keypad_update', 'status':{'alarm': True, 'alpha' : 'Panic Alarm'}},
    '626' : {'name':'PanicAlarmButtonOff', 'handler':'keypad_update', 'status':{'alarm': False, 'alpha' : 'Panic Alarm Cleared'}},
    '631' : {'name':'SmokeAlarmButton', 'handler':'keypad_update', 'status':{'alarm': True, 'alpha' : 'Smoke Alarm'}},
    '632' : {'name':'SmokeAlarmButtonOff', 'handler':'keypad_update', 'status':{'alarm': False, 'alpha' : 'Smoke Alarm Cleared'}},
    '660' : {'name':'PGMrelay', 'handler':'keypad_update', 'status':{'PGMrelay': True, 'alpha' : 'PGM output active'}},
    '800' : {'name':'LowBatTrouble', 'handler':'keypad_update', 'status':{'bat_trouble': True, 'alpha' : 'Low Battery'}},
    '801' : {'name':'LowBatTroubleOff', 'handler':'keypad_update', 'status':{'bat_trouble': False, 'alpha' : 'Low Battery Cleared'}},
    '802' : {'name':'ACTrouble', 'handler':'keypad_update', 'status':{'ac_present': False, 'alpha' : 'AC Power Lost'}},
    '803' : {'name':'ACTroubleOff', 'handler':'keypad_update', 'status':{'ac_present': True, 'alpha' : 'AC Power Restored'}},
    '829' : {'name':'SystemTamper', 'handler':'keypad_update', 'status':{'alpha' : 'System tamper'}},
    '830' : {'name':'SystemTamperOff', 'handler':'keypad_update', 'status':{'alpha' : 'System tamper Restored'}},
    '849' : {'name':'TroubleVerbose', 'handler':'keypad_update', 'status':None},

#ZONE BYPASS UPDATES
    '616' : {'name':'Zone Bypass', 'handler':'zone_bypass_update', 'status':None}
}

evl_verboseTrouble = {
 0 : 'Service is Required',
 1 : 'AC Power Lost',
 2 : 'Telephone Line Fault',
 3 : 'Failure to communicate',
 4 : 'Zone/Sensor Fault', 
 5 : 'Zone/Sensor Tamper',
 6 : 'Zone/Sensor Low Battery',
 7 : 'Loss of time'
}



#!/usr/bin/env python3
from setuptools import setup

setup(
    name="pyenvisalink",
    version="4.7",
    description=(
        "A python3 library for running asynchronus communications with envisalink "
        "alarm control panel modules."
    ),
    long_description=(
        "A python3 library for running asynchronus communications with envisalink "
        "alarm control panel modules."
    ),
    url="https://github.com/ufodone/pyenvisalink",
    author="David O'Neill",
    author_email="ufodone@gmail.com",
    license="MIT",
    packages=["pyenvisalink"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.10",
    ],
)



import logging
from pyenvisalink import HoneywellClient
from pyenvisalink import DSCClient
from pyenvisalink import AlarmState

_LOGGER = logging.getLogger(__name__)
COMMAND_ERR = "Cannot run this command while disconnected. Please run start() first."

class EnvisalinkAlarmPanel:
    """This class represents an envisalink-based alarm panel."""
        
    def __init__(self, host, port=4025, panelType='HONEYWELL',
                 envisalinkVersion=3, userName='user', password='user',
                 zoneTimerInterval=20, keepAliveInterval=30, eventLoop=None,
                 connectionTimeout=10, zoneBypassEnabled=False):
        self._host = host
        self._port = port
        self._connectionTimeout = connectionTimeout
        self._panelType = panelType
        self._evlVersion = envisalinkVersion
        self._username = userName
        self._password = password
        self._keepAliveInterval = keepAliveInterval
        self._zoneTimerInterval = zoneTimerInterval
        self._maxPartitions = 8
        if envisalinkVersion < 4:
            self._maxZones = 64
        else:
            self._maxZones = 128
        self._alarmState = AlarmState.get_initial_alarm_state(self._maxZones, self._maxPartitions)
        self._client = None
        self._eventLoop = eventLoop
        self._zoneBypassEnabled = zoneBypassEnabled
        
        self._loginSuccessCallback = self._defaultCallback
        self._loginFailureCallback = self._defaultCallback
        self._loginTimeoutCallback = self._defaultCallback
        self._commandResponseCallback = self._defaultCallback
        self._pollResponseCallback = self._defaultCallback
        self._keypadUpdateCallback = self._defaultCallback
        self._zoneStateChangeCallback = self._defaultCallback
        self._partitionStateChangeCallback = self._defaultCallback
        self._cidEventCallback = self._defaultCallback
        self._zoneTimerCallback = self._defaultCallback

        loggingconfig = {'level': 'DEBUG',
                     'format': '%(asctime)s %(levelname)s <%(name)s %(module)s %(funcName)s> %(message)s',
                     'datefmt': '%a, %d %b %Y %H:%M:%S'}

        logging.basicConfig(**loggingconfig)

    @property
    def host(self):
        return self._host
        
    @ property
    def port(self):
        return self._port

    @ property
    def connection_timeout(self):
        return self._connectionTimeout
        
    @property
    def user_name(self):
        return self._username
        
    @property
    def password(self):
        return self._password
        
    @property
    def panel_type(self):
        return self._panelType

    @property
    def envisalink_version(self):
        return self._evlVersion
        
    @property
    def keepalive_interval(self):
        return self._keepAliveInterval

    @property
    def zone_timer_interval(self):
        return self._zoneTimerInterval

    @property
    def alarm_state(self):
        return self._alarmState

    @property
    def callback_login(self):
        return self._defaultCallback
        
    @property
    def callback_login_success(self):
        return self._loginSuccessCallback

    @callback_login_success.setter
    def callback_login_success(self, value):
        self._loginSuccessCallback = value

    @property
    def callback_login_failure(self):
        return self._loginFailureCallback

    @callback_login_failure.setter
    def callback_login_failure(self, value):
        self._loginFailureCallback = value

    @property
    def callback_login_timeout(self):
        return self._loginTimeoutCallback

    @callback_login_timeout.setter
    def callback_login_timeout(self, value):
        self._loginTimeoutCallback = value

    @property
    def callback_poll_response(self):
        return self._pollResponseCallback

    @callback_poll_response.setter
    def callback_poll_response(self, value):
        self._pollResponseCallback = value

    @property
    def callback_command_response(self):
        return self._commandResponseCallback

    @callback_command_response.setter
    def callback_command_response(self, value):
        self._commandResponseCallback = value

    @property
    def callback_keypad_update(self):
        return self._keypadUpdateCallback

    @callback_keypad_update.setter
    def callback_keypad_update(self, value):
        self._keypadUpdateCallback = value

    @property
    def callback_zone_state_change(self):
        return self._zoneStateChangeCallback

    @callback_zone_state_change.setter
    def callback_zone_state_change(self, value):
        self._zoneStateChangeCallback = value

    @property
    def callback_partition_state_change(self):
        return self._partitionStateChangeCallback

    @callback_partition_state_change.setter
    def callback_partition_state_change(self, value):
        self._partitionStateChangeCallback = value

    @property
    def callback_realtime_cid_event(self):
        return self._cidEventCallback

    @callback_realtime_cid_event.setter
    def callback_realtime_cid_event(self, value):
        self._cidEventCallback = value

    @property
    def callback_zone_timer_dump(self):
        return self._zoneTimerCallback
 
    @callback_zone_timer_dump.setter
    def callback_zone_timer_dump(self, value):
        self._zoneTimerCallback = value
        
    def _defaultCallback(self, data):
        """This is the callback that occurs when the client doesn't subscribe."""
        _LOGGER.debug("Callback has not been set by client.")	    

    def start(self):
        """Connect to the envisalink, and listen for events to occur."""
        logging.info(str.format("Connecting to envisalink on host: {0}, port: {1}", self._host, self._port))
        if self._panelType == 'HONEYWELL':
            self._client = HoneywellClient(self, self._eventLoop)
            self._client.start()
        elif self._panelType == 'DSC':
            self._client = DSCClient(self, self._eventLoop)
            self._client.start()
        else:
            _LOGGER.error("Unexpected panel type.")    
        
    def stop(self):
        """Shut down and close our connection to the envisalink."""
        if self._client:
            _LOGGER.info("Disconnecting from the envisalink...")
            self._client.stop()
        else:
            _LOGGER.error(COMMAND_ERR)

    def dump_zone_timers(self):
        """Request a zone timer dump from the envisalink."""
        if self._client:
            self._client.dump_zone_timers()
        else:
            _LOGGER.error(COMMAND_ERR)

    def change_partition(self, partitionNumber):
        """Request that the default partition be changed."""
        if self._client:
            self._client.change_partition(partitionNumber)
        else:
            _LOGGER.error(COMMAND_ERR)

    def keypresses_to_default_partition(self, keypresses):
        """Send a key to the current partition."""
        if self._client:
            self._client.keypresses_to_default_partition(keypresses)
        else:
            _LOGGER.error(COMMAND_ERR)

    def keypresses_to_partition(self, partitionNumber, keypresses):
        """Send a key to a partition other than the current one."""
        if self._client:
            self._client.keypresses_to_partition(partitionNumber, keypresses)
        else:
            _LOGGER.error(COMMAND_ERR)

    def arm_stay_partition(self, code, partitionNumber):
        """Public method to arm/stay a partition."""
        if self._client:
            self._client.arm_stay_partition(code, partitionNumber)
        else:
            _LOGGER.error(COMMAND_ERR)

    def arm_away_partition(self, code, partitionNumber):
        """Public method to arm/away a partition."""
        if self._client:
            self._client.arm_away_partition(code, partitionNumber)
        else:
            _LOGGER.error(COMMAND_ERR)

    def arm_max_partition(self, code, partitionNumber):
        """Public method to arm/max a partition."""
        if self._client:
            self._client.arm_max_partition(code, partitionNumber)
        else:
            _LOGGER.error(COMMAND_ERR)

    def arm_night_partition(self, code, partitionNumber):
        """Public method to arm/night a partition."""
        if self._client:
            self._client.arm_night_partition(code, partitionNumber)
        else:
            _LOGGER.error(COMMAND_ERR)

    def disarm_partition(self, code, partitionNumber):
        """Public method to disarm a partition."""
        if self._client:
            self._client.disarm_partition(code, partitionNumber)
        else:
            _LOGGER.error(COMMAND_ERR)

    def panic_alarm(self, panic_type):
        """Public method to raise a panic alarm."""
        if self._client:
            self._client.panic_alarm(panic_type)
        else:
            _LOGGER.error(COMMAND_ERR)

    def toggle_zone_bypass(self, zone):
        """Public method to toggle a zone's bypass state."""
        if not self._zoneBypassEnabled:
            _LOGGER.error(COMMAND_ERR)
        elif self._client:
            self._client.toggle_zone_bypass(zone)
        else:
            _LOGGER.error(COMMAND_ERR)

    def command_output(self, code, partitionNumber, outputNumber):
        """Public method to activate an output"""
        if self._client:
            self._client.command_output(code, partitionNumber, outputNumber)
        else:
            _LOGGER.error(COMMAND_ERR)
